import asyncio
from typing import AsyncIterator, Awaitable, Callable, Dict

from pydantic import Field, validator
from rath.links.base import AsyncTerminatingLink
from rath.operation import GraphQLResult, Operation
from graphql import FieldNode, OperationType


def target_from_node(node: FieldNode) -> str:
    return (
        node.alias.value if hasattr(node, "alias") and node.alias else node.name.value
    )


class AsyncMockResolver:
    def to_dict(self):
        methods = [
            i for i in self.__class__.__dict__.keys() if i.startswith("resolve_")
        ]
        return {i[8:]: getattr(self, i) for i in methods}


class AsyncMockLink(AsyncTerminatingLink):
    query_resolver: Dict[str, Callable[[Operation], Awaitable[Dict]]] = Field(
        default_factory=dict, exclude=True
    )
    mutation_resolver: Dict[str, Callable[[Operation], Awaitable[Dict]]] = Field(
        default_factory=dict, exclude=True
    )
    subscription_resolver: Dict[str, Callable[[Operation], Awaitable[Dict]]] = Field(
        default_factory=dict, exclude=True
    )
    resolver: Dict[str, Callable[[Operation], Awaitable[Dict]]] = Field(
        default_factory=dict, exclude=True
    )

    @validator(
        "query_resolver",
        "mutation_resolver",
        "subscription_resolver",
        "resolver",
        pre=True,
    )
    @classmethod
    def coerce_resolver(cls, v):
        if isinstance(v, AsyncMockResolver):
            return v.to_dict()
        return v

    async def aexecute(self, operation: Operation) -> AsyncIterator[GraphQLResult]:

        if operation.node.operation == OperationType.QUERY:
            futures = []

            for op in operation.node.selection_set.selections:
                if op.name.value in self.query_resolver:
                    futures.append(self.query_resolver[op.name.value](operation))
                elif op.name.value in self.resolver:
                    futures.append(self.resolver[op.name.value](operation))
                else:
                    raise NotImplementedError(
                        f"Mocked Resolver for Query '{op.name.value}' not in resolvers: {self.query_resolver}, {self.resolver}  for AsyncMockLink"
                    )

            resolved = await asyncio.gather(*futures)
            yield GraphQLResult(
                data={
                    target_from_node(op): resolved[i]
                    for i, op in enumerate(operation.node.selection_set.selections)
                }
            )

        if operation.node.operation == OperationType.MUTATION:
            futures = []

            for op in operation.node.selection_set.selections:
                if op.name.value in self.mutation_resolver:
                    futures.append(self.mutation_resolver[op.name.value](operation))
                elif op.name.value in self.resolver:
                    futures.append(self.resolver[op.name.value](operation))
                else:
                    raise NotImplementedError(
                        f"Mocked Resolver for Query '{op.name.value}' not in resolvers: {self.mutation_resolver}, {self.resolver}  for AsyncMockLink"
                    )

            resolved = await asyncio.gather(*futures)
            yield GraphQLResult(
                data={
                    target_from_node(op): resolved[i]
                    for i, op in enumerate(operation.node.selection_set.selections)
                }
            )

        if operation.node.operation == OperationType.SUBSCRIPTION:
            futures = []
            assert (
                len(operation.node.selection_set.selections) == 1
            ), "Only one Subscription at a time possible"

            op = operation.node.selection_set.selections[0]
            if op.name.value in self.subscription_resolver:
                iterator = self.subscription_resolver[op.name.value](operation)
            elif op.name.value in self.resolver:
                iterator = self.resolver[op.name.value](operation)
            else:
                raise NotImplementedError(
                    f"Mocked Resolver for Query '{op.name.value}' not in resolvers: {self.subscription_resolver}, {self.resolver}  for AsyncMockLink"
                )

            async for event in iterator:
                yield GraphQLResult(data={target_from_node(op): event})

        else:
            raise NotImplementedError("Only subscription are mocked")



from typing import Any, Dict, Optional
from fakts.fakt.base import Fakt
from fakts.fakts import get_current_fakts
from rath.links.aiohttp import AIOHttpLink


class AioHttpConfig(Fakt):
    endpoint_url: str

    class Config:
        group = "aiohttp"


class FaktsAIOHttpLink(AIOHttpLink):
    endpoint_url: Optional[str]
    fakts_group: str
    fakt: Optional[AioHttpConfig]

    _old_fakt: Dict[str, Any] = None

    def configure(self, fakt: AioHttpConfig) -> None:
        self.endpoint_url = fakt.endpoint_url

    async def aconnect(self):
        fakts = get_current_fakts()

        if fakts.has_changed(self._old_fakt, self.fakts_group):
            self._old_fakt = await fakts.aget(self.fakts_group)
            self.configure(AioHttpConfig(**self._old_fakt))

        return await super().aconnect()



from asyncio.log import logger
from typing import Any, Callable, Dict
from graphql import (
    ListTypeNode,
    NamedTypeNode,
    NonNullTypeNode,
    OperationDefinitionNode,
    VariableNode,
)
from pydantic import BaseModel, Field
from rath.links.parsing import ParsingLink
from rath.operation import Operation
import logging

logger = logging.getLogger(__name__)


class TranspileHandler(BaseModel):
    """A Transpilation Handler

    A TranspileHandler is a function that takes any Any and returns a
    GraphQLType. It is used to implement custom type resolution.

    The default TranspileHandler is the identity function, which returns the
    type passed to it.
    """

    graphql_type: str
    name: str
    predicate: Callable[[Any], bool] = Field(exclude=True)
    parser: Callable[[Any], Any] = Field(exclude=True)

    class Config:
        arbitrary_types_allowed = True


class ListTranspileHandler(BaseModel):
    """A List Transpile Handler

    Similar to a TranspileHandler, but takes act on GraphqQLList Type of that type
    """

    graphql_type: str
    name: str
    predicate: Callable[[Any, int], bool] = Field(exclude=True)
    parser: Callable[[Any, int], Any] = Field(exclude=True)

    class Config:
        arbitrary_types_allowed = True


class TranspilationError(Exception):
    """A transpilation Exception"""


class TranspilationHandlerException(TranspilationError):
    """A transpilation Exception happening within a TranspileHandler"""


class TranspileRegistry(BaseModel):
    """A Registry to hold TranspileHandlers"""

    list_handlers: Dict[str, Dict[str, ListTranspileHandler]] = Field(
        default_factory=dict
    )
    item_handlers: Dict[str, Dict[str, TranspileHandler]] = Field(default_factory=dict)

    def register_item(
        self,
        graphql_type: str,
        predicate: Callable[[Any], bool],
        name=None,
    ):
        """A Decorator for registering a TranspileHandler

        If acting on a List of this type, the handle_list parameter should be set to True.

        Example:
            ```python
            registry = TranspileRegistry()

            @registry.register_item("ID", lambda x: isinstance(x, BaseModel))
            def transpile_id_to_id(x):
                return str(x.id)
            ```


        Args:
            graphql_type (str): The graphql Type to act upon
            predicate (Callable[[Any], bool]): A predicate Function
            name (_type_, optional): A name for this hanlder. Defaults to the function name.
        """

        def decorator(func):
            """The decorator function

            Args:
                func (_type_): The function to use as a handler

            Returns:
                Callable: The function
            """

            x = name or func.__name__

            self.item_handlers.setdefault(graphql_type, {})[x] = TranspileHandler(
                predicate=predicate, parser=func, name=x, graphql_type=graphql_type
            )
            return func

        return decorator

    def register_list(
        self,
        graphql_type: str,
        predicate: Callable[[Any, str], bool],
        name=None,
    ):
        """A Decorator for registering a TranspileHandler

        If acting on a List of this type, the handle_list parameter should be set to True.

        Example:
            ```python
            registry = TranspileRegistry()

            @registry.register_list("InputVector", lambda x, listdepth: isinstance(x, np.ndarray))
            def transpile_numpy_array_to_vectors(x, listdepth):
                assert listdepth == 1, "Only one level of nesting is supported"
                return [InputVector(x) for x in x]
            ```
        Args:
            graphql_type (str): The graphql Type to act upon
            predicate (Callable[[Any], bool]): A predicate Function
            handle_list (bool, optional): Should we act on lists of this type. Defaults to False.
            name (_type_, optional): A name for this hanlder. Defaults to the function name.
        """

        def decorator(func):
            """The decorator function

            Args:
                func (_type_): The function to use as a handler

            Returns:
                Callable: The function
            """

            x = name or func.__name__

            self.list_handlers.setdefault(graphql_type, {})[x] = ListTranspileHandler(
                predicate=predicate, parser=func, name=x, graphql_type=graphql_type
            )

            return func

        return decorator


def recurse_transpile(
    key,
    var: VariableNode,
    value: Any,
    registry: TranspileRegistry,
    in_list=0,
    strict=False,
):
    """Recurse Transpile a variable according to a registry and
    its definition

    Args:
        key (_type_): The key of the variable
        var (VariableNode): The variable definition node correspoin to this variable
        value (Any): The to transpile valued
        registry (TranspileRegistry): The transpile registry to use
        in_list (bool, optional): Recursive Parameter. That will be set to the list depth. Defaults to False.
        strict (bool, optional): Should we error on predicate errors. Defaults to False.

    Raises:
        TranspilationError: _description_

    Returns:
        Any: The transpiled value or the original value if no handler matched
    """

    if isinstance(var, NonNullTypeNode):
        return recurse_transpile(
            key, var.type, value, registry, in_list=in_list, strict=strict
        )
    if isinstance(var, ListTypeNode):
        return recurse_transpile(
            key, var.type, value, registry, in_list=in_list + 1, strict=strict
        )
    if isinstance(var, NamedTypeNode):
        try:
            if in_list > 0:
                if var.name.value in registry.list_handlers:
                    for k, handler in registry.list_handlers[var.name.value].items():
                        try:
                            predicate = handler.predicate(value, in_list)
                        except Exception as e:
                            if strict:
                                raise TranspilationHandlerException(
                                    f"Handler {handler} predicate failed"
                                ) from e
                            logger.warning(
                                f"Handler {handler} failed on predicating {value}. Please check your predicate for edge cases"
                            )
                            continue

                        if predicate:
                            parsed_value = handler.parser(value, in_list)
                            assert (
                                parsed_value is not None
                            ), f"Handler {handler} failed on parsing {value}. Please check your parser for edge cases"
                            return parsed_value

                return value

            else:
                if var.name.value in registry.item_handlers:
                    for k, handler in registry.item_handlers[var.name.value].items():
                        try:
                            predicate = handler.predicate(value)
                        except Exception as e:
                            if strict:
                                raise Exception(f"Handler {handler} failed with {e}")
                            logger.warning(
                                f"Handler {handler} failed on predicating {value}. Please check your predicate for edge cases"
                            )
                            continue
                        if predicate:
                            parsed_value = [handler.parser(value) for value in value]
                            assert (
                                parsed_value is not None
                            ), f"Handler {handler} failed on parsing {value}. Please check your parser for edge cases"
                            return parsed_value

                else:
                    return value

        except Exception as e:
            raise TranspilationError(
                f"Transpilation for variable {key} on graphql type `{var.name.value}` ! Original value: {repr(value)}"
            ) from e


def transpile(
    op: OperationDefinitionNode,
    variables: Dict[str, Any],
    registry: TranspileRegistry,
    strict=False,
) -> Dict[str, Any]:
    """Transpile

    Transpiles a operations variabels to a dictionary of variables, with
    json serializable values according to a transpile registry.

    Args:
        op (OperationDefinitionNode): The operation definition node,
        registry (TranspileRegistry): The registry
        strict (bool, optional): Should we fail if a handler predicate fails. Defaults to False.

    Returns:
        Dict: The transpiled variables
    """
    variable_nodes = {
        var_node.variable.name.value: var_node.type
        for var_node in op.variable_definitions
        if var_node.variable.name.value in variables
    }

    transpiled_variables = {
        key: recurse_transpile(key, variable, variables[key], registry, strict=strict)
        for key, variable in variable_nodes.items()
    }

    return transpiled_variables


class TranspileLink(ParsingLink):
    """Transpile Link

    Transpile Link is a link that transpiles variables according to a transpile registry.


    Attributes:
        registry (TranspileRegistry): The transpile registry to use
        strict (bool): Should we fail if a handler predicate fails. Defaults to False.

    """

    registry: TranspileRegistry
    strict: bool = False

    async def aparse(self, operation: Operation) -> Operation:
        operation.variables = transpile(
            operation.node, operation.variables, self.registry, self.strict
        )
        return operation



from koil.composition import KoiledModel
from pydantic import Field
from rath.errors import NotConnectedError, NotEnteredError
from rath.links.base import TerminatingLink
from typing import (
    AsyncIterator,
    Dict,
    Any,
    Iterator,
    Optional,
)
from rath.operation import GraphQLResult, opify
from contextvars import ContextVar
from koil import unkoil_gen, unkoil


current_rath = ContextVar("current_rath")


class WrappedContext:
    def __init__(self, value):
        self.__value: ContextVar = value

    def set(self, value):
        return self.__value.set(value)

    def reset(self, token):
        self.__value.reset(token)
        return

    @classmethod
    def __get_validators__(cls):
        # one or more validators may be yielded which will be called in the
        # order to validate the input, each validator will receive as an input
        # the value returned from the previous validator
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if isinstance(v, ContextVar):
            return cls(v)
        raise TypeError("Needs to be either a instance of ContextVar or a string")


class Rath(KoiledModel):
    link: Optional[TerminatingLink] = None
    context: Optional[WrappedContext] = Field(
        default_factory=lambda: WrappedContext(current_rath)
    )

    _connected = False
    _entered = False
    auto_connect = True
    connect_on_enter: bool = False

    async def aconnect(self):
        if not self._entered:
            raise NotEnteredError(
                "You need to use enter `async with Rath(...) as rath`"
            )
        await self.link.aconnect()
        self._connected = True

    async def adisconnect(self):
        await self.link.adisconnect()
        self._connected = False

    async def aquery(
        self,
        query: str,
        variables: Dict[str, Any] = None,
        headers: Dict[str, Any] = {},
        operation_name=None,
        **kwargs,
    ) -> GraphQLResult:
        if not self._connected:
            if not self.auto_connect:
                raise NotConnectedError(
                    "Rath is not connected. Please use `async with Rath(..., auto_connect=True) as rath` or use `await rath.aconnect() before`"
                )
            await self.aconnect()

        op = opify(query, variables, headers, operation_name, **kwargs)

        async for data in self.link.aexecute(op):
            return data

    def query(
        self,
        query: str,
        variables: Dict[str, Any] = None,
        headers: Dict[str, Any] = {},
        operation_name=None,
        **kwargs,
    ) -> GraphQLResult:
        return unkoil(self.aquery, query, variables, headers, operation_name, **kwargs)

    def subscribe(
        self,
        query: str,
        variables: Dict[str, Any] = None,
        headers: Dict[str, Any] = {},
        operation_name=None,
        **kwargs,
    ) -> Iterator[GraphQLResult]:
        return unkoil_gen(
            self.asubscribe, query, variables, headers, operation_name, **kwargs
        )

    async def asubscribe(
        self,
        query: str,
        variables: Dict[str, Any] = None,
        headers: Dict[str, Any] = {},
        operation_name=None,
        **kwargs,
    ) -> AsyncIterator[GraphQLResult]:
        if not self._connected:
            if not self.auto_connect:
                raise NotConnectedError(
                    "Rath is not connected. Please use `async with Rath(..., auto_connect=True) as rath` or use `await rath.aconnect() before`"
                )
            await self.aconnect()

        op = opify(query, variables, headers, operation_name, **kwargs)
        async for data in self.link.aexecute(op):
            yield data

    async def __aenter__(self):
        self._entered = True
        if self.context:
            self.context.set(self)
        await self.link.__aenter__()
        return self

    async def __aexit__(self, *args, **kwargs):
        if self._connected:
            await self.adisconnect()

        await self.link.__aexit__(*args, **kwargs)
        self._entered = False
        if self.context:
            self.context.set(None)

    class Config:
        underscore_attrs_are_private = True



from typing import AsyncIterator, Awaitable, Callable, Optional

from pydantic import Field
from rath.links.base import ContinuationLink
from rath.operation import GraphQLResult, Operation
from rath.links.errors import AuthenticationError


async def fake_loader():
    raise Exception("No Token loader specified")


class AuthTokenLink(ContinuationLink):

    token_loader: Callable[[], Awaitable[str]]
    token_refresher: Optional[Callable[[], Awaitable[str]]]

    maximum_refresh_attempts: int = 3
    load_token_on_connect: bool = True
    load_token_on_enter: bool = True

    _token: str = None

    async def aconnect(self):
        if self.load_token_on_connect:
            await self.token_loader()

    async def aexecute(
        self, operation: Operation, retry=0, **kwargs
    ) -> AsyncIterator[GraphQLResult]:

        operation.context.headers[
            "Authorization"
        ] = f"Bearer {await self.token_loader()}"
        try:

            async for result in self.next.aexecute(operation, **kwargs):
                yield result

        except AuthenticationError as e:
            retry = retry + 1
            operation.context.headers[
                "Authorization"
            ] = f"Bearer {await self.token_refresher()}"
            if retry > self.maximum_refresh_attempts:
                raise AuthenticationError("Maximum refresh attempts reached") from e

            async for result in self.aexecute(operation, retry=retry + 1, **kwargs):
                yield result

    class Config:
        underscore_attrs_are_private = True
        arbitary_types_allowed = True



from typing import Generic, Type, TypeVar

from koil.qt import QtRunner
from qtpy import QtCore
from rath import Rath

T = TypeVar("T")


class QtRathQuery(QtRunner, Generic[T]):
    started: QtCore.Signal = QtCore.Signal()
    errored: QtCore.Signal = QtCore.Signal(Exception)
    cancelled: QtCore.Signal = QtCore.Signal()
    returned: QtCore.Signal = QtCore.Signal(object)

    def __init__(self, operation: Type[T], rath: Rath):
        async def coro(*args, **kwargs):
            assert rath is not None, "No rath found"
            print("RUnning")

            result = await rath.aquery(operation.Meta.document, *args, **kwargs)
            return operation(**result.data)

        super().__init__(coro)
        self.operation = operation

    def run(self, **kwargs):
        arg = self.operation.Arguments(**kwargs).dict()
        return super().run(arg)



from typing import Dict, Any, Callable

from rath.operation import Operation


def recurse_parse_variables(
    variables: Dict,
    predicate: Callable[[str, Any], bool],
    apply: Callable[[Any], Any],
) -> Dict:
    """Parse Variables

    Recursively traverse variables, applying the apply function to the value if the predicate
    returns True.

    Args:
        variables (Dict): The dictionary to parse.
        predicate (Callable[[str, Any], bool]):The path this is in
        apply (Callable[[Any], Any]): _description_

    Returns:
        Dict: _description_
    """

    def recurse_extract(obj, path: str = None):
        """
        recursively traverse obj, doing a deepcopy, but
        replacing any file-like objects with nulls and
        shunting the originals off to the side.
        """

        if isinstance(obj, list):
            nulled_obj = []
            for key, value in enumerate(obj):
                value = recurse_extract(
                    value,
                    f"{path}.{key}" if path else key,
                )
                nulled_obj.append(value)
            return nulled_obj
        elif isinstance(obj, dict):
            nulled_obj = {}
            for key, value in obj.items():
                value = recurse_extract(value, f"{path}.{key}" if path else key)
                nulled_obj[key] = value
            return nulled_obj
        elif predicate(path, obj):
            return apply(obj)
        else:
            return obj

    dicted_variables = recurse_extract(variables)

    return dicted_variables


def recurse_parse_variables_with_operation(
    variables: Dict,
    operation: Operation,
    predicate: Callable[[str, Any], bool],
    apply: Callable[[Any], Any],
) -> Dict:
    """Parse Variables

    Recursively traverse variables, applying the apply function to the value if the predicate
    returns True.

    Args:
        variables (Dict): The dictionary to parse.
        predicate (Callable[[str, Any], bool]): _description_
        apply (Callable[[Any], Any]): _description_

    Returns:
        Dict: _description_
    """

    def recurse_extract(obj, path: str = None):
        """
        recursively traverse obj, doing a deepcopy, but
        replacing any file-like objects with nulls and
        shunting the originals off to the side.
        """

        if isinstance(obj, list):
            nulled_obj = []
            for key, value in enumerate(obj):
                value = recurse_extract(
                    value,
                    f"{path}.{key}" if path else key,
                )
                nulled_obj.append(value)
            return nulled_obj
        elif isinstance(obj, dict):
            nulled_obj = {}
            for key, value in obj.items():
                value = recurse_extract(value, f"{path}.{key}" if path else key)
                nulled_obj[key] = value
            return nulled_obj
        elif predicate(path, obj):
            return apply(obj)
        else:
            return obj

    dicted_variables = recurse_extract(variables)

    return dicted_variables



from rath.operation import Operation
from rath.links.parsing import ParsingLink


class ShrinkByID:
    id: str

    async def ashrink(self):
        return self.id

    def shrink(self):
        return self.id


class ShrinkingLink(ParsingLink):
    async def aparse(self, operation: Operation) -> Operation:
        shrinked_variables = {
            key: await var.ashrink()
            for key, var in operation.variables.items()
            if hasattr(var, "ashrink")
        }
        operation.variables.update(shrinked_variables)
        return operation



from pydantic import BaseModel


class ID(str):
    """A custom scalar for IDs. If passed a pydantic model it an id property
    it will automatically validate to that models id"""

    def __init__(self, value) -> None:
        self.value = value

    @classmethod
    def __get_validators__(cls):
        # one or more validators may be yielded which will be called in the
        # order to validate the input, each validator will receive as an input
        # the value returned from the previous validator
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if isinstance(v, BaseModel):
            if hasattr(v, "id"):
                return cls(v.id)
            else:
                raise TypeError("This needs to be a instance of BaseModel with an id")

        if isinstance(v, str):
            return cls(v)

        if isinstance(v, int):
            return cls(str(v))

        raise TypeError(
            "Needs to be either a instance of BaseModel (with an id) or a string"
        )



from typing import AsyncIterator, Optional
from koil.composition import KoiledModel
from rath.operation import GraphQLResult, Operation


class Link(KoiledModel):
    async def aconnect(self):
        pass

    async def adisconnect(self):
        pass

    async def __aenter__(self) -> None:
        pass

    async def __aexit__(self, *args, **kwargs) -> None:
        pass

    def aexecute(self, operation: Operation, **kwargs) -> AsyncIterator[GraphQLResult]:
        raise NotImplementedError(
            f"Please overwrite the asubscribe method in {self.__class__.__name__}"
        )


class TerminatingLink(Link):
    pass


class AsyncTerminatingLink(TerminatingLink):
    """An Async Termination Link

    This is a link that terminates the query or subscription, aka
    it is a link that does not have a next link. It normally
    is used to make network requests with the already parsed
    operations

    Args:
        TerminatingLink (_type_): _description_
    """

    async def aexecute(self, operation: Operation) -> AsyncIterator[GraphQLResult]:
        raise NotImplementedError("Your Async Transport needs to overwrite this method")


class ContinuationLink(Link):
    next: Optional[Link] = None

    def set_next(self, next: Link):
        self.next = next

    async def aexecute(self, operation: Operation, **kwargs) -> GraphQLResult:
        async for x in self.next.aexecute(operation, **kwargs):
            yield x



from typing import Dict
from pydantic import BaseModel
from rath.operation import Operation
from rath.links.parsing import ParsingLink


def parse_variables(
    variables: Dict,
) -> Dict:
    def recurse_extract(path, obj):
        """
        recursively traverse obj, doing a deepcopy, but
        replacing any file-like objects with nulls and
        shunting the originals off to the side.
        """

        if isinstance(obj, list):
            nulled_obj = []
            for key, value in enumerate(obj):
                value = recurse_extract(f"{path}.{key}", value)
                nulled_obj.append(value)
            return nulled_obj
        elif isinstance(obj, dict):
            nulled_obj = {}
            for key, value in obj.items():
                value = recurse_extract(f"{path}.{key}", value)
                nulled_obj[key] = value
            return nulled_obj
        elif isinstance(obj, BaseModel):
            return obj.dict(by_alias=True)
        else:
            # base case: pass through unchanged
            return obj

    dicted_variables = recurse_extract("variables", variables)

    return dicted_variables


class DictingLink(ParsingLink):
    async def aparse(self, operation: Operation) -> Operation:
        shrinked_variables = parse_variables(operation.variables)
        operation.variables.update(shrinked_variables)
        return operation



from typing import List

from pydantic import validator, root_validator
from rath.links.base import ContinuationLink, Link, TerminatingLink
from rath.operation import Operation


class ComposedLink(TerminatingLink):
    links: List[Link]

    @validator("links")
    def validate(cls, value):
        if not value:
            raise ValueError("ComposedLink requires at least one link")

        if not all(isinstance(link, Link) for link in value):
            raise ValueError("ComposedLink requires all links to be Links")

        assert isinstance(
            value[-1], TerminatingLink
        ), "Last link must be a TerminatingLink"
        for link in value[:-1]:
            assert isinstance(
                link, ContinuationLink
            ), f"All but the last must be ContinuationLinks: check {link}"

        return value

    async def aconnect(self):
        for link in self.links:
            await link.aconnect()

    async def adisconnect(self):
        for link in self.links:
            await link.adisconnect()

    async def __aenter__(self):
        for i in range(len(self.links) - 1):
            self.links[i].set_next(self.links[i + 1])

        for link in self.links:
            await link.__aenter__()

    async def __aexit__(self, *args, **kwargs):
        for link in self.links:
            await link.__aexit__(*args, **kwargs)

    async def aexecute(self, operation: Operation, **kwargs):
        async for result in self.links[0].aexecute(operation):
            yield result


class TypedComposedLink(TerminatingLink):
    _firstlink: Link = None

    async def aconnect(self):
        for key, link in self:
            if isinstance(link, Link):
                await link.aconnect()

    async def adisconnect(self):
        for key, link in self:
            if isinstance(link, Link):
                await link.adisconnect()

    async def __aenter__(self):

        current_link = None

        for key, link in self:
            if isinstance(link, Link):
                if current_link:
                    current_link.set_next(link)
                    await current_link.__aenter__()
                else:
                    self._firstlink = link
                    await self._firstlink.__aenter__()

                current_link = link

        await current_link.__aenter__()

    async def __aexit__(self, *args, **kwargs):
        for key, link in self:
            if isinstance(link, Link):
                await link.__aexit__(*args, **kwargs)

    async def aexecute(self, operation: Operation, **kwargs):

        async for result in self._firstlink.aexecute(operation):
            yield result

    class Config:
        underscore_attrs_are_private = True


def compose(*links: Link) -> ComposedLink:
    """
    Composes a list of Links into a single Link.
    """

    return ComposedLink(links=links)



from typing import List

from graphql import OperationType
from pydantic import Field
from turms.plugins.funcs import (
    FunctionDefinition,
    Kwarg,
    FuncsPlugin,
    FuncsPluginConfig,
)


class RathPluginConfig(FuncsPluginConfig):
    definitions: List[FunctionDefinition] = [
        FunctionDefinition(type=OperationType.MUTATION, use="rath.turms.funcs.execute"),
        FunctionDefinition(
            type=OperationType.MUTATION,
            is_async=True,
            use="rath.turms.funcs.aexecute",
        ),
        FunctionDefinition(type=OperationType.QUERY, use="rath.turms.funcs.execute"),
        FunctionDefinition(
            is_async=True,
            type=OperationType.QUERY,
            use="rath.turms.funcs.aexecute",
        ),
        FunctionDefinition(
            type=OperationType.SUBSCRIPTION, use="rath.turms.funcs.subscribe"
        ),
        FunctionDefinition(
            is_async=True,
            type=OperationType.SUBSCRIPTION,
            use="rath.turms.funcs.asubscribe",
        ),
    ]
    extra_kwargs: List[Kwarg] = [
        Kwarg(
            key="rath",
            type="rath.Rath",
            description="The rath client to use",
            default=None,
        ),
    ]


class RathFuncsPlugin(FuncsPlugin):
    config: RathPluginConfig = Field(default_factory=RathPluginConfig)



from rath.links.base import ContinuationLink
from rath.operation import Operation
from rath.operation import Operation

import io
import aiohttp
from typing import AsyncGenerator


FILE_CLASSES = (
    io.IOBase,
    aiohttp.StreamReader,
    AsyncGenerator,
)


from typing import Any, Dict, Tuple, Type


def parse_variables(
    variables: Dict,
    file_classes: Tuple[Type[Any], ...] = FILE_CLASSES,
) -> Tuple[Dict, Dict]:
    files = {}

    def recurse_extract(path, obj):
        """
        recursively traverse obj, doing a deepcopy, but
        replacing any file-like objects with nulls and
        shunting the originals off to the side.
        """
        nonlocal files

        if isinstance(obj, list):
            nulled_obj = []
            for key, value in enumerate(obj):
                value = recurse_extract(f"{path}.{key}", value)
                nulled_obj.append(value)
            return nulled_obj
        elif isinstance(obj, dict):
            nulled_obj = {}
            for key, value in obj.items():
                value = recurse_extract(f"{path}.{key}", value)
                nulled_obj[key] = value
            return nulled_obj
        elif isinstance(obj, file_classes):
            # extract obj from its parent and put it into files instead.
            files[path] = obj
            return None

        elif hasattr(obj, "__file__"):
            # extract obj from its parent and put it into files instead.
            files[path] = obj.value
            return None

        else:
            # base case: pass through unchanged
            return obj

    nulled_variables = recurse_extract("variables", variables)

    return nulled_variables, files


class FileExtraction(ContinuationLink):
    async def aexecute(self, operation: Operation) -> Operation:
        operation.variables, operation.context.files = parse_variables(
            operation.variables
        )

        async for result in self.next.aexecute(operation):
            yield result



from typing import Any, Dict, Optional
from fakts.fakt.base import Fakt
from fakts.fakts import get_current_fakts
from rath.links.aiohttp import AIOHttpLink
from rath.links.auth import AuthTokenLink
from rath.links.websockets import WebSocketLink
from herre import current_herre


class WebsocketHttpConfig(Fakt):
    ws_endpoint_url: str

    class Config:
        group = "aiohttp"


class FaktsWebsocketLink(WebSocketLink):
    ws_endpoint_url: Optional[str]
    fakts_group: str = "websocket"

    _old_fakt: Dict[str, Any] = {}

    def configure(self, fakt: WebsocketHttpConfig) -> None:
        self.ws_endpoint_url = fakt.ws_endpoint_url
        self.token_loader = current_herre.get().aget_token

    async def aconnect(self):
        fakts = get_current_fakts()

        if fakts.has_changed(self._old_fakt, self.fakts_group):
            self._old_fakt = await fakts.aget(self.fakts_group)
            self.configure(WebsocketHttpConfig(**self._old_fakt))

        return await super().aconnect()



from typing import Awaitable, Callable, Dict, Optional
from graphql import OperationType
from pydantic import Field
import websockets
import json
import asyncio
from websockets.exceptions import (
    ConnectionClosedError,
)
import logging
import uuid
from rath.links.errors import LinkNotConnectedError, TerminatingLinkError

from rath.operation import (
    GraphQLException,
    GraphQLResult,
    Operation,
    SubscriptionDisconnect,
)
from rath.links.base import AsyncTerminatingLink


logger = logging.getLogger(__name__)


GQL_WS_SUBPROTOCOL = "graphql-ws"

# all the message types
GQL_CONNECTION_INIT = "connection_init"
GQL_START = "start"
GQL_STOP = "stop"
GQL_CONNECTION_TERMINATE = "connection_terminate"
GQL_CONNECTION_ERROR = "connection_error"
GQL_CONNECTION_ACK = "connection_ack"
GQL_DATA = "data"
GQL_ERROR = "error"
GQL_COMPLETE = "complete"
GQL_CONNECTION_KEEP_ALIVE = "ka"

WEBSOCKET_DEAD = "websocket_dead"
WEBSOCKET_CANCELLED = "websocket_cancelled"


class CorrectableConnectionFail(TerminatingLinkError):
    pass


class DefiniteConnectionFail(TerminatingLinkError):
    pass


class InvalidPayload(TerminatingLinkError):
    pass


async def none_token_loader():
    raise Exception("Token loader was not set")


class WebSocketLink(AsyncTerminatingLink):
    ws_endpoint_url: str
    allow_reconnect: bool = True
    time_between_retries = 4
    max_retries = 3
    token_loader: Callable[[], Awaitable[str]] = Field(
        default_factory=lambda: none_token_loader, exclude=True
    )

    _connected: bool = False
    _alive: bool = False
    _send_queue: asyncio.Queue = None
    _connection_task: asyncio.Task = None
    _ongoing_subscriptions: Optional[Dict[str, asyncio.Queue]] = None

    async def aforward(self, message):
        await self._send_queue.put(message)

    async def __aenter__(self):
        self._ongoing_subscriptions = {}
        self._send_queue = asyncio.Queue()

    async def aconnect(self):
        logger.info("Connecting Websockets")
        self._connection_task = asyncio.create_task(self.websocket_loop())
        self._connected = True

    async def adisconnect(self):
        self._connection_task.cancel()

        try:
            await self._connection_task
        except asyncio.CancelledError:
            logger.info(f"Websocket Transport {self} succesfully disconnected")

    async def __aexit__(self, *args, **kwargs):
        if self._connection_task:
            await self.adisconnect()

    async def websocket_loop(self, retry=0):
        send_task = None
        receive_task = None
        try:
            try:
                token = await self.token_loader()
                url = (
                    f"{self.ws_endpoint_url}?token={token}"
                    if token
                    else self.ws_endpoint_url
                )
                async with websockets.connect(
                    url,
                    subprotocols=[GQL_WS_SUBPROTOCOL],
                ) as client:
                    logger.info("Websocket successfully connected")

                    send_task = asyncio.create_task(self.sending(client))
                    receive_task = asyncio.create_task(self.receiving(client))

                    self._alive = True
                    done, pending = await asyncio.wait(
                        [send_task, receive_task],
                        return_when=asyncio.FIRST_EXCEPTION,
                    )
                    self._alive = False

                    for task in pending:
                        task.cancel()

                    for task in done:
                        raise task.exception()

            except Exception as e:
                logger.warning("Websocket excepted. Trying to recover", exc_info=True)
                raise CorrectableConnectionFail from e

        except CorrectableConnectionFail as e:
            logger.info(
                f"Trying to Recover from Exception {e} Reconnect is {self.allow_reconnect} Retry: {retry}"
            )
            if retry > self.max_retries or not self.allow_reconnect:
                logger.error("Max retries reached. Aborting")
                raise DefiniteConnectionFail("Exceeded Number of Retries")

            await asyncio.sleep(self.time_between_retries)
            logger.info(f"Retrying to connect")
            await self.broadcast({"type": WEBSOCKET_DEAD, "error": e})
            await self.websocket_loop(retry=retry + 1)

        except DefiniteConnectionFail as e:
            logger.error("Websocket excepted closed definetely", exc_info=True)
            self.connection_dead = True
            raise e

        except asyncio.CancelledError as e:
            logger.info("Websocket got cancelled. Trying to shutdown graceully")
            if send_task and receive_task:
                send_task.cancel()
                receive_task.cancel()

                cancellation = await asyncio.gather(
                    send_task, receive_task, return_exceptions=True
                )
            raise e

        except Exception as e:
            logger.error("Websocket excepted", exc_info=True)
            self.connection_dead = True
            raise e

    async def sending(self, client, headers=None):
        payload = {"type": GQL_CONNECTION_INIT, "payload": {"headers": headers}}
        await client.send(json.dumps(payload))

        try:
            while True:
                message = await self._send_queue.get()
                logger.debug("GraphQL Websocket: >>>>>> " + message)
                await client.send(message)
                self._send_queue.task_done()
        except asyncio.CancelledError as e:
            logger.debug("Sending Task sucessfully Cancelled")  #
            raise e

    async def receiving(self, client):
        try:
            async for message in client:
                logger.debug("GraphQL Websocket: <<<<<<< " + message)
                try:
                    message = json.loads(message)
                    await self.broadcast(message)
                except json.JSONDecodeError as err:
                    logger.warning(
                        "Ignoring. Server sent invalid JSON data: %s \n %s",
                        message,
                        err,
                    )
        except Exception as e:
            logger.warning("Websocket excepted. Trying to recover", exc_info=True)
            raise e

    async def broadcast(self, message: dict):
        type = message["type"]

        if type == GQL_CONNECTION_KEEP_ALIVE:
            return

        if type == GQL_CONNECTION_KEEP_ALIVE:
            return

        if type == WEBSOCKET_DEAD:
            # notify all subscriptipns that the websocket is dead
            for subscription in self._ongoing_subscriptions.values():
                await subscription.put(message)
            return

        if type in [GQL_DATA, GQL_COMPLETE]:

            if "id" not in message:
                raise InvalidPayload(f"Protocol Violation. Expected 'id' in {message}")

            id = message["id"]
            assert (
                id in self._ongoing_subscriptions
            ), "Received Result for subscription that is no longer or was never active"
            await self._ongoing_subscriptions[id].put(message)

    async def aexecute(self, operation: Operation):
        if self._connection_task == False:
            raise LinkNotConnectedError("Websocket_link is not connected")

        assert (
            operation.node.operation == OperationType.SUBSCRIPTION
        ), "Operation is not a subscription"
        assert not operation.context.files, "We cannot send files through websockets"

        id = operation.id
        subscribe_queue = asyncio.Queue()
        self._ongoing_subscriptions[id] = subscribe_queue

        payload = {
            "headers": operation.context.headers,
            "query": operation.document,
            "variables": operation.variables,
        }
        frame = {"id": id, "type": GQL_START, "payload": payload}
        await self.aforward(json.dumps(frame))

        while True:
            answer = await subscribe_queue.get()

            if answer["type"] == GQL_DATA:
                payload = answer["payload"]

                if "errors" in payload:
                    raise GraphQLException(
                        "\n".join([e["message"] for e in payload["errors"]])
                    )

                if "data" in payload:
                    yield GraphQLResult(data=payload["data"])
                    subscribe_queue.task_done()

            if answer["type"] == WEBSOCKET_DEAD:
                raise SubscriptionDisconnect(
                    f"Subcription {id} failed propagating Error {operation}"
                )

            if answer["type"] == GQL_COMPLETE:
                logger.info(f"Subcription done {operation}")
                return

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True



from typing import Callable

from pydantic import Field
from rath.operation import Operation
from rath.links.base import TerminatingLink


class SplitLink(TerminatingLink):
    left: TerminatingLink
    right: TerminatingLink
    split: Callable[[Operation], bool] = Field(exclude=True)

    async def aexecute(self, operation: Operation, **kwargs) -> Operation:
        iterator = (
            self.left.aexecute(operation, **kwargs)
            if self.split(operation)
            else self.right.aexecute(operation, **kwargs)
        )

        async for res in iterator:
            yield res

    async def aconnect(self):
        await self.left.aconnect()
        await self.right.aconnect()

    async def adisconnect(self):
        await self.left.adisconnect()
        await self.right.adisconnect()

    async def __aenter__(self):
        await self.left.__aenter__()
        await self.right.__aenter__()

    async def __aexit__(self, *args, **kwargs):
        await self.left.__aexit__(*args, **kwargs)
        await self.right.__aexit__(*args, **kwargs)


def split(
    left: TerminatingLink, right: TerminatingLink, split: Callable[[Operation], bool]
):
    """
    Splits a Link into two paths. Acording to a predicate function. If predicate returns
    true, the operation is sent to the left path, otherwise to the right path.
    """
    return SplitLink(left=left, right=right, split=split)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rath',
 'rath.contrib.fakts',
 'rath.contrib.fakts.links',
 'rath.contrib.herre.links',
 'rath.links',
 'rath.links.testing',
 'rath.qt',
 'rath.turms',
 'rath.turms.plugins']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0',
 'graphql-core>=3.2.0,<4.0.0',
 'koil==0.2.6',
 'pydantic>=1.9.0,<2.0.0']

extras_require = \
{'websockets': ['websockets>=10.2,<11.0']}

setup_kwargs = {
    'name': 'rath',
    'version': '0.2.8',
    'description': 'aiohttp powered apollo like graphql client',
    'long_description': '# rath\n\n[![codecov](https://codecov.io/gh/jhnnsrs/rath/branch/master/graph/badge.svg?token=UGXEA2THBV)](https://codecov.io/gh/jhnnsrs/rath)\n[![PyPI version](https://badge.fury.io/py/rath.svg)](https://pypi.org/project/rath/)\n[![Maintenance](https://img.shields.io/badge/Maintained%3F-yes-green.svg)](https://pypi.org/project/rath/)\n![Maintainer](https://img.shields.io/badge/maintainer-jhnnsrs-blue)\n[![PyPI pyversions](https://img.shields.io/pypi/pyversions/rath.svg)](https://pypi.python.org/pypi/rath/)\n[![PyPI status](https://img.shields.io/pypi/status/rath.svg)](https://pypi.python.org/pypi/rath/)\n[![PyPI download month](https://img.shields.io/pypi/dm/rath.svg)](https://pypi.python.org/pypi/rath/)\n\n### BETA\n\n## Inspiration\n\nRath is like Apollo, but for python. It adheres to the design principle of Links and enables complex GraphQL\nsetups, like seperation of query and subscription endpoints, dynamic token loading, etc..\n\n## Installation\n\n```bash\npip install rath\n```\n\n## Usage Example\n\n```python\nfrom rath.links.auth import AuthTokenLink\nfrom rath.links.aiohttp import AIOHttpLink\nfrom rath.links import compose, split\nfrom rath.gql import gql\n\nasync def aload_token():\n    return "SERVER_TOKEN"\n\n\nauth = AuthTokenLink(token_loader=aload_token)\nlink = AIOHttpLink(endpoint_url="https://api.spacex.land/graphql/")\n\n\nwith Rath(links=compose(auth,link)) as rath:\n    query = """query TestQuery {\n      capsules {\n        id\n        missions {\n          flight\n        }\n      }\n    }\n    """\n\n    result = rath.query(query)\n```\n\nThis example composes both the AuthToken and AioHttp link: During each query the Bearer headers are set to the retrieved token, on authentication fail (for example if Token Expired) the\nAuthToken automatically refetches the token and retries the query.\n\n## Async Usage\n\nRath is build for async usage but uses koil, for async/sync compatibility\n\n```python\nfrom rath.links.auth import AuthTokenLink\nfrom rath.links.aiohttp import AIOHttpLink\nfrom rath.links import compose, split\nfrom rath.gql import gql\n\nasync def aload_token():\n    return "SERVER_TOKEN"\n\n\nauth = AuthTokenLink(token_loader=aload_token)\nlink = AIOHttpLink(endpoint_url="https://api.spacex.land/graphql/")\n\n\nasync def main():\n\n  async with Rath(links=compose(auth,link)) as rath:\n      query = """query TestQuery {\n        capsules {\n          id\n          missions {\n            flight\n          }\n        }\n      }\n      """\n\n      result = await rath.query(query)\n\nasyncio.run(main())\n```\n\n## Example Transport Switch\n\nLinks allow the composition of additional logic based on your graphql operation. For example you might want\nto use different grapqhl transports for different kind of operations (e.g using websockets for subscriptions,\nbut using standard http requests for potential caching on queries and mutations). This can be easily\naccomplished by providing a split link.\n\n```python\nlink = SplitLink(\n  AioHttpLink(url="https://api.spacex.land/graphql/"),\n  WebsocketLink(url="ws://api.spacex.land/graphql/",\n  lamda o: o.node.operation == OperationType.SUBSCRIPTION\n)\n\nrath = Rath(link=link)\n\n```\n\n## Included Links\n\n- Validating Link (validate query against local schema (or introspect the schema))\n- Reconnecting WebsocketLink\n- AioHttpLink (supports multipart uploads)\n- SplitLink (allows to split the terminating link - Subscription into WebsocketLink, Query, Mutation into Aiohttp)\n- AuthTokenLink (Token insertion with automatic refres\n\n## Authentication\n\nIf you want to use rath with herre for getting access_tokens in oauth2/openid-connect scenarios, there is also a herre link\nin this repository\n\n### Why Rath\n\nWell "apollo" is already taken as a name, and rath (according to wikipedia) is an etruscan deity identified with Apollo.\n\n## Rath + Turms\n\nRath works especially well with turms generated typed operations:\n\n```python\nimport asyncio\nfrom examples.api.schema import aget_capsules\nfrom rath.rath import Rath\nfrom rath.links.aiohttp import AIOHttpLink\nfrom rath.links.auth import AuthTokenLink\nfrom rath.links.compose import compose\n\n\nasync def token_loader():\n    return ""\n\n\nlink = compose(\n    AuthTokenLink(token_loader), AIOHttpLink("https://api.spacex.land/graphql/")\n)\n\n\nrath = Rath(\n    link=link,\n    register=True, # allows global access (singleton-antipattern, but rath has no state)\n)\n\n\nasync def main():\n\n    async with rath:\n        capsules = await aget_capsules() # fully typed pydantic powered dataclasses generated through turms\n        print(capsules)\n\n\nasyncio.run(main())\n\n```\n\n## Examples\n\nThis github repository also contains an example client with a turms generated query with the public SpaceX api, as well as a sample of the generated api.\n',
    'author': 'jhnnsrs',
    'author_email': 'jhnnsrs@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import Awaitable, Callable, Optional
from herre.herre import Herre
from rath.links.aiohttp import AIOHttpLink
from rath.links.auth import AuthTokenLink
from herre import current_herre


class HerreAuthLink(AuthTokenLink):
    herre: Optional[Herre]
    token_loader: Optional[Callable[[], Awaitable[str]]]
    token_refresher: Optional[Callable[[], Awaitable[str]]]

    async def aconnect(self):
        self.herre = self.herre or current_herre.get()
        self.token_loader = self.herre.aget_token
        self.token_refresher = self.herre.arefresh_token
        return await super().aconnect()



class LinkError(Exception):
    """Base class for all transport errors."""



class LinkNotConnectedError(Exception):
    pass

    def __init__(self, message) -> None:
        super().__init__(
            message
            + "\n To connect link please use either an async or sync context manager "
        )


class TerminatingLinkError(LinkError):
    """Raised when a terminating link is called."""



class ContinuationLinkError(LinkError):
    pass


class AuthenticationError(TerminatingLinkError):
    pass



from http import HTTPStatus
import json
from typing import Any, Dict, List

import aiohttp
from graphql import OperationType
from pydantic import Field
from rath.operation import GraphQLException, GraphQLResult, Operation
from rath.links.base import AsyncTerminatingLink
from rath.links.errors import AuthenticationError
import logging

logger = logging.getLogger(__name__)


class AIOHttpLink(AsyncTerminatingLink):
    endpoint_url: str
    auth_errors: List[HTTPStatus] = Field(
        default_factory=lambda: (HTTPStatus.FORBIDDEN,)
    )

    _session = None

    async def __aenter__(self) -> None:
        self._session = await aiohttp.ClientSession().__aenter__()

    async def __aexit__(self, *args, **kwargs) -> None:
        await self._session.__aexit__(*args, **kwargs)

    async def aexecute(self, operation: Operation) -> GraphQLResult:
        payload = {"query": operation.document}

        if operation.node.operation == OperationType.SUBSCRIPTION:
            raise NotImplementedError(
                "Aiohttp Transport does not support subscriptions"
            )

        if len(operation.context.files.items()) > 0:
            payload["variables"] = operation.variables

            files = operation.context.files
            data = aiohttp.FormData()

            file_map = {str(i): [path] for i, path in enumerate(files)}

            # Enumerate the file streams
            # Will generate something like {'0': <_io.BufferedReader ...>}
            file_streams = {str(i): files[path] for i, path in enumerate(files)}
            operations_str = json.dumps(payload)

            data.add_field(
                "operations", operations_str, content_type="application/json"
            )
            file_map_str = json.dumps(file_map)
            data.add_field("map", file_map_str, content_type="application/json")

            for k, v in file_streams.items():
                data.add_field(
                    k,
                    v,
                    filename=getattr(v, "name", k),
                )

            post_kwargs: Dict[str, Any] = {"data": data}

        else:
            payload["variables"] = operation.variables
            post_kwargs = {"json": payload}

        async with aiohttp.ClientSession() as session:
            async with session.post(
                self.endpoint_url, headers=operation.context.headers, **post_kwargs
            ) as response:

                if response.status == HTTPStatus.OK:
                    result = await response.json()

                if response.status in self.auth_errors:
                    raise AuthenticationError(
                        f"Token Expired Error {operation.context.headers}"
                    )

                json_response = await response.json()

                if "errors" in json_response:
                    raise GraphQLException(
                        "\n".join([e["message"] for e in json_response["errors"]])
                    )

                if "data" not in json_response:

                    raise Exception(f"Response does not contain data {json_response}")

                yield GraphQLResult(data=json_response["data"])

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True
